unit OpenALObjectU;

interface

uses
  openal, SysUtils;

type
  TOpenALObject = class(TObject)
    private
      fBuffer: TALuint;
      fLoop: TALint;
      fSource: TALuint;
    protected
      function Get_IsPlaying: Boolean;
    public
      constructor Create;
      destructor Destroy; override;
      procedure Load; virtual; abstract;
      procedure Update;
      procedure Play;
      procedure Pause;
      procedure Stop;
      property IsPlaying: Boolean read Get_IsPlaying;
    end;

  TOpenALObjectFromFile = class(TOpenALObject)
    private
      fFilename: string;
    public
      constructor Create(aFilename: string = '');
      procedure Load; override;
  end;

implementation

{ TOpenALObject }

constructor TOpenALObject.Create;
var
  lError: TALenum;
begin
  alGetError;
  alGenBuffers(1, @fBuffer);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('alGenBuffers error: %s',
      [alGetString(lError)]);
  end;
  alGenSources(1, @fSource);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('alGenSources error: %s',
      [alGetString(lError)]);
  end;
end;

destructor TOpenALObject.Destroy;
var
  lError: TALenum;
begin
  if IsPlaying then
  begin
    Stop;
  end;
  alGetError;
  alDeleteSources(1, @fSource);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('alDeleteSources error: %s',
      [alGetString(lError)]);
  end;
  alDeleteBuffers(1, @fBuffer);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('alDeleteBuffers error: %s',
      [alGetString(lError)]);
  end;
  inherited;
end;

function TOpenALObject.Get_IsPlaying: Boolean;
var
  lCurrentState: TALenum;
  lError: TALenum;
begin
  // Check whether the source is playing and set status.
  alGetError;
  alGetSourcei(fSource, AL_SOURCE_STATE, @lCurrentState);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('AL_SOURCE_STATE error %s',
      [alGetString(lError)]);
  end;
  Result := (lCurrentState = AL_PLAYING);
end;

procedure TOpenALObject.Pause;
begin
  alSourcePause(fSource);
end;

procedure TOpenALObject.Play;
begin
  alSourcePlay(fSource);
end;

procedure TOpenALObject.Stop;
begin
  alSourceStop(fSource);
end;

procedure TOpenALObject.Update;
begin
  // Pass updated values to OpenAL
  alSourcei(fSource, AL_LOOPING, fLoop);
end;

{ TOpenALObjectFromFile }

constructor TOpenALObjectFromFile.Create(aFilename: string);
begin
  inherited Create;
  fFilename := aFilename;
end;

procedure TOpenALObjectFromFile.Load;
var
  lFormat: TALEnum;
  lSize: TALSizei;
  lFrequency: TALSizei;
  lLoop: TALInt;
  lData: TALVoid;
begin
  //load data from the file - assumes it's a .wav file
  alutLoadWAVFile(fFilename, lFormat, lData, lSize, lFrequency, lLoop);
  //assign this data to our buffer
  alBufferData(fBuffer, lFormat, lData, lSize, lFrequency);
  if alGetError <> AL_NO_ERROR then
    raise Exception.Create('Cannot assign wave data to buffer');
  //remove the wavedata from memory
  AlutUnloadWav(lFormat, lData, lSize, lFrequency);
  fLoop := lLoop;
  //assign our buffer to the source
  AlSourcei ( fSource, AL_BUFFER, fBuffer);
end;

end.
