unit OpenALStreamingObjectU;

interface

uses
  Classes, ExtCtrls, openal;

const
  MAX_BUFFERS     = 4;
  BUFFER_SIZE     = 4096;
  SAMPLE_RATE     = 44100;
  TWO_PI          = 6.2831853;

type
  TDataContainer = class(TObject)
  private
    fDataBuffer: TALuint;
    fPcmBuffer: Array[0..BUFFER_SIZE-1] of SmallInt;
    function GetPcmBuffer(Index: Integer): SmallInt;
    function GetPcmBufferPtr: Pointer;
    procedure SetPcmBuffer(Index: Integer; const Value: SmallInt);
  public
    constructor Create;
    destructor Destroy; override;
    property DataBuffer: TALuint read fDataBuffer write fDataBuffer;
    property PcmBuffer[Index: Integer]: SmallInt read GetPcmBuffer write SetPcmBuffer;
    property PcmBufferPtr: Pointer read GetPcmBufferPtr;
  end;

  TOpenALStreamingObject = class(TObject)
  private
    fAvailableBuffers: TList;
    fCurrentSampleIndex: Int64;
    fQueuedBuffers: TList;
    fSource: TALuint;
    fTimer: TTimer;
    fBufferArray: Array[0..MAX_BUFFERS-1] of TALUint;
    fFrequency: Integer;
    fVolume: Integer;
    procedure CreateOpenAL;
    procedure DestroyOpenAL;
    procedure QueueNextBuffer;
    procedure UnqueueBuffers;
    function Get_IsPlaying: Boolean;
  protected

  public
    constructor Create;
    destructor Destroy; override;
    procedure Play; virtual;
    procedure Stop;
    procedure Update(Sender: TObject);
    property IsPlaying: Boolean read Get_IsPlaying;
    property Frequency: Integer read fFrequency write fFrequency;
    property Volume: Integer read fVolume write fVolume;
  end;

  // Audio file description
  TAudioFileDescription = record
    SampleRate:       Double;
    FormatID:         Cardinal;
    FormatFlags:      Cardinal;
    BytesPerPacket:   Cardinal;
    FramesPerPacket:  Cardinal;
    BytesPerFrame:    Cardinal;
    ChannelsPerFrame: Cardinal;
    BitsPerChannel:   Cardinal;
    AudioFormat:      Integer;
    AudioDataSize:    Cardinal;
  end;

implementation

uses
  SysUtils;

{ TOpenALStreamingObject }

constructor TOpenALStreamingObject.Create;
begin
  inherited;
  fCurrentSampleIndex := 1;
  fFrequency := 440;
  fVolume := 20000;
  fAvailableBuffers := TList.Create;
  fAvailableBuffers.Capacity := MAX_BUFFERS;
  fQueuedBuffers := TList.Create;
  fQueuedBuffers.Capacity := MAX_BUFFERS;
  CreateOpenAL;
  // Create the timer to update our stream
  fTimer := TTimer.Create(nil);
  fTimer.Name := 'StreamingTimer';
  fTimer.Enabled := False;
  fTimer.OnTimer := Update;
  fTimer.Interval := 10;
end;

procedure TOpenALStreamingObject.CreateOpenAL;
var
  lError: TALenum;
  lIndex: Integer;
  lContainer: TDataContainer;
begin
  alGetError;
  alGenBuffers(MAX_BUFFERS, @fBufferArray);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('alGenBuffers error: %s', [alGetString(lError)]);
  end;
  alGenSources(1, @fSource);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('alGenSources error: %s', [alGetString(lError)]);
  end;
  for lIndex := 0 to MAX_BUFFERS - 1 do
  begin
    lContainer := TDataContainer.Create;
    lContainer.DataBuffer := fBufferArray[lIndex];
    fAvailableBuffers.Add(lContainer);
  end;
end;

destructor TOpenALStreamingObject.Destroy;
var
  lIndex: Integer;
  lDataContainer: TDataContainer;
begin
  DestroyOpenAL;
  for lIndex := 0 to fQueuedBuffers.Count - 1 do
  begin
    lDataContainer := fQueuedBuffers[lIndex];
    lDataContainer.Free;
  end;
  for lIndex := 0 to fAvailableBuffers.Count - 1 do
  begin
    lDataContainer := fAvailableBuffers[lIndex];
    lDataContainer.Free;
  end;
  fTimer.Free;
  inherited;
end;

procedure TOpenALStreamingObject.DestroyOpenAL;
var
  lError: TALenum;
begin
  alGetError;
  alDeleteSources(1, @fSource);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('alDeleteSources error: %s',
      [alGetString(lError)]);
  end;
  alDeleteBuffers(MAX_BUFFERS, @fBufferArray);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('alDeleteBuffers error: %s',
      [alGetString(lError)]);
  end;
end;

function TOpenALStreamingObject.Get_IsPlaying: Boolean;
var
  lCurrentState: TALenum;
  lError: TALenum;
begin
  // Check whether the source is playing and set status.
  alGetError;
  alGetSourcei(fSource, AL_SOURCE_STATE, @lCurrentState);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('AL_SOURCE_STATE error %s', [alGetString(lError)]);
  end;
  Result := (lCurrentState = AL_PLAYING);
end;

procedure TOpenALStreamingObject.Play;
begin
  fTimer.Enabled := True;
  alSourcePlay(fSource);
end;

procedure TOpenALStreamingObject.QueueNextBuffer;
var
  lBufferSize: Integer;
  lCount: Integer;
  lCurrentBuffer: TALuint;
  lDataContainer: TDataContainer;
  lError: TALenum;
  lIndex: Integer;
  lFormat: TALenum;
  lQueuedBuffers: TALint;
  lSampleRate: Integer;
  lSampleValue: Double;
begin
  lQueuedBuffers := 0;
  alGetError;
  alGetSourcei(fSource, AL_BUFFERS_QUEUED, @lQueuedBuffers);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('AL_BUFFERS_QUEUED error: %s',
      [alGetString(lError)]);
  end;
  lDataContainer := fAvailableBuffers.Last;
  lCurrentBuffer := lDataContainer.DataBuffer;
  lCount := Round(BUFFER_SIZE / SizeOf(Int16));
  for lIndex := 0 to lCount - 1 do
  begin
    // Basic sine wave formula
    // x = a * sin(2*Pi*f*h*t) + p
    // where
    //  a = amplitude     (fixed to 1000.0)
    //  f = frequency     (fixed to Middle C i.e. 261.626Hz
    //  h = harmonic      (fixed to 1.0/2.0)
    //  t = time interval
    //  p = phase offset  (fixed at 0.0)
    lSampleValue := fVolume *
      Sin(TWO_PI * fFrequency * fCurrentSampleIndex / (2.0 * SAMPLE_RATE)) + 0.0;
    Inc(fCurrentSampleIndex);
    lDataContainer.PcmBuffer[lIndex] := Round(lSampleValue);
  end;
  lSampleRate := SAMPLE_RATE;
  lBufferSize := BUFFER_SIZE;
  lFormat := AL_FORMAT_MONO16;
  alBufferData(lDataContainer.DataBuffer, lFormat,
    lDataContainer.PcmBufferPtr, lBufferSize, lSampleRate);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('alBufferData error: %s',
      [alGetString(lError)]);
  end;
  alSourceQueueBuffers(fSource, 1, @lCurrentBuffer);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('alSourceQueueBuffers error: %s',
      [alGetString(lError)]);
  end;
  fQueuedBuffers.Insert(0, fAvailableBuffers.Last);
  fAvailableBuffers.Delete(fAvailableBuffers.IndexOf(fAvailableBuffers.Last));
end;

procedure TOpenALStreamingObject.Stop;
begin
  fTimer.Enabled := False;
  alSourceStop(fSource);
end;

procedure TOpenALStreamingObject.UnqueueBuffers;
var
  lError: TALenum;
  lBuffersProcessed: TALint;
  lUnqueuedBuffer: TALuint;
begin
  lBuffersProcessed := 0;
  alGetError;
  alGetSourcei(fSource, AL_BUFFERS_PROCESSED, @lBuffersProcessed);
  lError := alGetError;
  if (lError <> AL_NO_ERROR) then
  begin
    raise Exception.CreateFmt('AL_BUFFERS_PROCESSED error: %s',
          [alGetString(lError)]);
  end;

  while (lBuffersProcessed > 0) do
  begin
    alSourceUnqueueBuffers(fSource, 1, @lUnqueuedBuffer);
    lError := alGetError;
    if (lError <> AL_NO_ERROR) then
    begin
      raise Exception.CreateFmt('alSourceUnqueueBuffers error: %s',
        [alGetString(lError)]);
    end;
    if (fQueuedBuffers.Count > 0) then
    begin
      fAvailableBuffers.Add(fQueuedBuffers.Last);
      fQueuedBuffers.Delete(fQueuedBuffers.IndexOf(fQueuedBuffers.Last));
    end;
    Dec(lBuffersProcessed);
  end;
end;

procedure TOpenALStreamingObject.Update(Sender: TObject);
var
  lCurrentState: TALenum;
  lError: TALenum;
  lQueuedBuffers: TALint;
begin
  UnqueueBuffers;
  if fAvailableBuffers.Count > 0 then
  begin
    QueueNextBuffer;
  end else begin
    alGetSourcei(fSource, AL_SOURCE_STATE, @lCurrentState);
    lError := alGetError;
    if (lError <> AL_NO_ERROR) then
    begin
      raise Exception.CreateFmt('AL_SOURCE_STATE error: %s',
        [alGetString(lError)]);
    end;
    if (lCurrentState <> AL_PLAYING) then
    begin
      lQueuedBuffers := 0;
      alGetSourcei(fSource, AL_BUFFERS_QUEUED, @lQueuedBuffers);
      lError := alGetError;
      if (lError <> AL_NO_ERROR) then
      begin
        raise Exception.CreateFmt('AL_BUFFERS_QUEUED error: %s',
          [alGetString(lError)]);
      end;
      if (lQueuedBuffers > 0) then
      begin
        alSourcePlay(fSource);
        lError := alGetError;
        if (lError <> AL_NO_ERROR) then
        begin
          raise Exception.CreateFmt('alSourcePlay error: %s',
            [alGetString(lError)]);
        end;
      end;
    end;
  end;
end;

{ TOpenALStreamFromFile }

{ TDataContainer }

constructor TDataContainer.Create;
begin
  fDataBuffer := 0;
end;

destructor TDataContainer.Destroy;
begin
  inherited;
end;

function TDataContainer.GetPcmBuffer(Index: Integer): SmallInt;
begin
  Result := fPcmBuffer[Index];
end;

function TDataContainer.GetPcmBufferPtr: Pointer;
begin
  Result := @fPcmBuffer;
end;

procedure TDataContainer.SetPcmBuffer(Index: Integer; const Value: SmallInt);
begin
  fPcmBuffer[Index] := Value;
end;

end.
